﻿(function () {

    // String

    String.prototype.format = function () {
        var args = arguments;
        return this.replace(/%((%)|[0-9]+)/g, function (m, num, pct) {
            return pct || (num = parseInt(num, 10), num >= 0 && num < args.length ? args[num] : m);
        });
    };
    String.prototype.trimMillisecondsFromIsoDate = function () {
        //this solves a bug with IE
        return new Date(Date.parse(this.substring(0, 23) + (this.indexOf('+') >= 0 ? this.substring(this.indexOf('+')) : '')));
    };

    // Date

    var dateMarkers = {
        d: ['getDate', function (v) { return ("0" + v).substr(-2, 2); } ],
        m: ['getMonth', function (v) { return ("0" + v).substr(-2, 2); } ],
        n: ['getMonth', function (v) {
            var mthNames = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"];
            return mthNames[v];
        } ],
        w: ['getDay', function (v) {
            var dayNames = ["sun", "mon", "tue", "wed", "thu", "fri", "sat"];
            return dayNames[v];
        } ],
        y: ['getFullYear'],
        H: ['getHours', function (v) { return ("0" + v).substr(-2, 2); } ],
        M: ['getMinutes', function (v) { return ("0" + v).substr(-2, 2); } ],
        S: ['getSeconds', function (v) { return ("0" + v).substr(-2, 2); } ],
        i: ['toISOString', null]
    };

    var format = function (date, fmt) {
        var dateTxt = fmt.replace(/%(.)/g, function (m, p) {
            var rv = date[(dateMarkers[p])[0]]();

            if (dateMarkers[p][1] != null)
                rv = dateMarkers[p][1](rv);

            return rv;
        });

        return dateTxt;
    };

    Date.prototype.format = function (fmt) {
        return format(this, fmt);
    };


    // handling browser focus
    var hidden, change, vis = {
        hidden: "visibilitychange",
        mozHidden: "mozvisibilitychange",
        webkitHidden: "webkitvisibilitychange",
        msHidden: "msvisibilitychange",
        oHidden: "ovisibilitychange" /* not currently supported */
    };
    for (hidden in vis) {
        if (vis.hasOwnProperty(hidden) && hidden in document) {
            change = vis[hidden];
            break;
        }
    }
    if (change)
        document.addEventListener(change, onchange);
    else if ($.browser.version.indexOf('Internet Explorer') >= 0) // l 9 and lower
        document.onfocusin = document.onfocusout = onchange;
    else
        window.onfocus = window.onblur = onchange;
    function onchange(evt) {
        evt = evt || window.event;
        elmahr.setBrowserVisible(evt.type !== "blur" && evt.type !== "focusout"
                                ? !this[hidden]
                                : false);
    }

    // ====================
    // elmahr global module
    // ====================

    this.elmahr = (function () {

        var play = 'play',
            pause = 'pause',
            show = 'show',
            hide = 'hide';

        // ===========
        // view models
        // ===========

        function ErrorViewModel(envelope, sourceId, infoUrl) {

            if (!(this instanceof ErrorViewModel)) {
                return new ErrorViewModel(envelope, sourceId, infoUrl);
            }

            var self = this;

            var time = envelope.Error.Time.trimMillisecondsFromIsoDate().toLocaleString(),
                   e = envelope.Error;

            self.id = e.Id;
            self.applicationName = envelope.ApplicationName;
            self.sourceId = sourceId;
            self.isFull = e.IsFull;
            self.infoUrl = infoUrl;
            self.message = e.Message;
            self.time = time;
            self.rawTime = envelope.Error.Time;
            self.isoTime = e.IsoTime;
            self.host = e.Host;
            self.type = e.Type;
            self.shortType = e.ShortType;
            self.source = e.Source;
            self.detail = e.Detail;
            self.user = e.User;
            self.statusCode = e.StatusCode;
            self.webHostHtmlMessage = e.WebHostHtmlMessage;
            self.hasYsod = e.HasYsod;
            self.url = e.Url;
            self.browserSupportUrl = elmahr.config.imagesPath + e.BrowserSupportUrl;
            self.reconnectClass = envelope.Class !== null;

            self.serverVariables = [];
            self.form = [];
            self.cookies = [];

            for (var sv in e.ServerVariables) {
                self.serverVariables.push({
                    'key': sv,
                    'value': e.ServerVariables[sv]
                });
            }
            
            for (var f in e.Form) {
                self.form.push({
                    'key': f,
                    'value': e.Form[f]
                });
            }
            
            for (var c in e.Cookies) {
                self.cookies.push({
                    'key': c,
                    'value': e.Cookies[c]
                });
            }

            self.popup = function (target) {
                if (target.isFull) {
                    selected(target);
                    elmahr.onSelectError(target);
                }
                else
                    elmahr.selectTarget(target.id, target.sourceId);
            };
        }

        function ErrorBarrierViewModel(application, when, offset, countBefore, countAfter) {

            if (!(this instanceof ErrorBarrierViewModel)) {
                return new ErrorBarrierViewModel(application, when, offset, countBefore, countAfter);
            }

            var self = this;

            self.application = application;
            self.when = when.trimMillisecondsFromIsoDate().format('%d-%n-%y');
            self.offset = offset ? -1 * offset + ' days before latest day with errors' : 'latest day with errors';
            self.countBefore = countBefore;
            self.countAfter = countAfter;

            var clear = function (target, event, action) {
                if (application.deletor())
                    return;
                elmahr.toggleBinLoading($(event.target), true);
                action(function () {
                    elmahr.toggleBinLoading($(event.target), false);
                });
            };

            self.clearErrorsBefore = function (target, event) {
                clear(target, event, function (post) { elmahr.askToClearErrorsBefore(application.sourceId, when, post); });
            };
            self.clearErrorsAfter = function (target, event) {
                clear(target, event, function (post) { elmahr.askToClearErrorsAfter(application.sourceId, when, post); });
            };
        }

        function ApplicationViewModel(applicationName, infoUrl, sourceId, hasTestExceptionUrl, boxColor, index) {

            if (!(this instanceof ApplicationViewModel)) {
                return new ApplicationViewModel(applicationName, infoUrl, sourceId, hasTestExceptionUrl, boxColor, index);
            }

            var self = this;

            self.applicationName = applicationName;
            self.infoUrl = infoUrl;
            self.sourceId = sourceId;
            self.index = index;
            self.hasTestExceptionUrl = hasTestExceptionUrl;
            self.boxColorBase = boxColor
                              ? boxColor
                              : 'boxColor' + (index % 3);

            //observables

            self.boxColor = ko.observable(self.boxColorBase + ' ' + (elmahr.isApplicationBlocked(sourceId) ? 'grayed' : ''));
            self.errors = ko.observableArray([]);
            self.lastError = ko.observable(null);
            self.errorsCount = ko.observable(0);
            self.errorsStats = ko.observableArray([]);
            self.errorsStatsCount = ko.observable(0);
            self.errorsStatsDate = ko.observable(null);
            self.errorsStatsFormattedDate = ko.observable(null);
            self.deletor = ko.observable(null);

            //functions

            self.addError = function (envelope, append) {
                var model = new ErrorViewModel(envelope, self.sourceId, infoUrl);
                if (append)
                    self.errors.push(model);
                else
                    self.errors.unshift(model);
            };

            self.feedTitle = function () {
                return elmahr.isApplicationBlocked(sourceId)
                       ? elmahr.config.messageFeedPaused
                       : elmahr.config.messageFeedRunning;
            };

            self.toggleApplication = function (target, e) {
                var t = $(e.currentTarget);
                var c = t.attr('class');
                var n = c === play ? pause : play;

                t.switchClass(c, n, 0);

                elmahr.toggleApplication(target.sourceId, n);

                self.boxColor(self.boxColorBase + ' ' + (c === play ? 'grayed' : ''));

                t.attr('title', self.feedTitle());
            };

            self.toggleApplicationErrors = function (target, e) {
                var t = $(e.currentTarget);
                var c = t.attr('class');
                var n = c === show ? hide : show;

                t.switchClass(c, n, 0);

                $('#' + elmahr.config.appBoxIdPrefix + self.index).toggle('fast');
            };

            self.raiseTestException = function (target) {
                elmahr.raiseTestException(target.sourceId);
            };

            self.getErrorsResume = function (target) {
                elmahr.getErrorsResume(target.sourceId);
            };

            var clearer = function (when, predicate) {
                var tbr = [],
                    errs = self.errors();

                for (var i = 0; i < errs.length; i++)
                    if (predicate(errs[i].rawTime, when))
                        tbr.push(errs[i]);

                for (var j = 0; j < tbr.length; j++)
                    self.errors.remove(tbr[j]);

                self.errorsCount(self.errorsCount() - tbr.length >= 0
                                 ? self.errorsCount() - tbr.length
                                 : 0);

                elmahr.clearErrors(when, predicate, self.sourceId);
            };

            self.askToClearAllErrors = function (target, event) {
                if (self.deletor())
                    return;
                elmahr.toggleBinLoading($(event.target), true);
                self.deletor(
                    { 'action':
                        function () {
                            elmahr.clearAllErrors(self.sourceId, function () {
                                elmahr.toggleBinLoading($(event.target), false);
                            });
                            self.deletor(null);
                        }
                    });
            };

            self.askToClearErrorsBefore = function (when, onFail) {
                self.deletor({ 'when': when, 'action': function () { elmahr.clearErrorsBefore(self.sourceId, when, onFail); self.deletor(null); } });
            };

            self.askToClearErrorsAfter = function (when, onFail) {
                self.deletor({ 'when': when, 'action': function () { elmahr.clearErrorsAfter(self.sourceId, when, onFail); self.deletor(null); } });
            };

            self.clearAllErrors = function () {
                clearer(null, function () { return true; });
            };

            self.clearErrorsBefore = function (when) {
                clearer(when, function (l, r) { return l < r; });
            };

            self.clearErrorsAfter = function (when) {
                clearer(when, function (l, r) { return l >= r; });
            };

            self.cancelDeletor = function () {
                self.deletor(null);
                elmahr.toggleBinLoading($("a.loading"), false);
            };

            //computed observables

            self.applicationClass = ko.computed(function () {
                return elmahr.isApplicationBlocked(sourceId) ? pause : play;
            }, this);

            self.titlePlayPause = ko.computed(self.feedTitle, this);

            self.updateErrorsStats = function (resume) {

                self.errorsStats([]);
                self.errorsStatsCount(0);

                if (resume) {
                    var barriers = resume.Barriers;

                    self.errorsStatsCount(resume.Count);
                    self.errorsStatsDate(resume.When);
                    self.errorsStatsFormattedDate(resume.When.trimMillisecondsFromIsoDate().format('%d-%n-%y'));

                    for (var i = 0; i < barriers.length; i++)
                        self.errorsStats.push(new ErrorBarrierViewModel(self,
                            barriers[i].When,
                            barriers[i].Offset,
                            barriers[i].CountBefore,
                            barriers[i].CountAfter));

                }
                selectedApp(self);
            };
        }

        // ==========
        // local vars
        // ==========

        var elmahrConnector = $.connection.elmahr,
            selected = ko.observable(null),
            selectedApp = ko.observable(null),
            latest = ko.observable(null),

            popupDisabled = false,

            blockedErrorTypes = [],
            isErrorTypeBlocked = function (type) {
                return $.inArray(type, blockedErrorTypes) > -1;
            },

            blockedApplications = [],
            isApplicationBlocked = function (sourceId) {
                return $.inArray(sourceId, blockedApplications) > -1;
            },

            lowestIndex = -1,
            updateLowestIndex = function (e, append) {
                lowestIndex = append
                            ? e.id
                            : lowestIndex === -1
                              ? e.id
                              : lowestIndex;
            },

            hookPlugins = function (plugins) {
                if ($.isEmptyObject(plugins))
                    return;
                for (var i = 0; i < plugins.length; i++) {
                    var plugin = plugins[i];
                    var selector =
                        plugin.Key.toLowerCase() === 'script' || plugin.Key.toLowerCase() === 'css'
                        ? 'head'
                        : '#' + plugin.Key;
                    $(selector).append(plugin.Value);
                }
            },

            clearer = function (sourceId, when, remoteClearer, localClearer) {
                var apps = elmahr.data.advertisedApplications();
                remoteClearer().done(function () {
                    for (var a = 0; a < apps.length; a++) {
                        var app = apps[a];
                        if (apps[a].sourceId === sourceId) {
                            localClearer(app);
                        }
                    }
                    elmahr.getErrorsResume(sourceId);
                    elmahr.refreshStats();
                });
            },

            askToClear = function (sourceId, action) {
                var apps = elmahr.data.advertisedApplications();
                for (var a = 0; a < apps.length; a++) {
                    var app = apps[a];
                    if (app.sourceId === sourceId)
                        action(app);
                }
            };

        // =============
        // elmahr object
        // =============

        return {

            extras: {},

            plugins: {},

            data: {
                advertisedApplications: ko.observableArray([]),
                allErrors: ko.observableArray([]),
                unseenErrors: 'no',
                selected: selected,
                selectedApp: selectedApp,
                latest: latest,
                ready: false,
                total: ko.observable(0)
            },

            config: {
                root: '/',
                imagesPath: 'content/images/',
                title: 'ElmahR Dashboard',
                appBoxIdPrefix: 'app',
                messageFeedPaused: 'This feed is paused, press here to receive errors',
                messageFeedRunning: 'This feed is running, press here to pause it'
            },

            //default functions meant to be overridden

            onSelectError: function (error) { },
            onNewError: function () { },
            onRememberMe: function (remember) { },
            hideAskForMore: function () { },
            statusConnecting: function () { },
            statusConnected: function () { },
            statusSendingCommand: function () { },
            statusStartingHubs: function () { },
            statusShowUI: function () { },
            statusRefreshingStats: function () { },
            statusReceiving: function () { },
            startTransfer: function () { },
            stopTransfer: function () { },
            showMessage: function (text, severity) { },
            getSortedApplications: function () { },
            toggleBinLoading: function (loading) { },
            onGetErrorsResumeDone: function (loading) { },

            // methods

            disablePopup: function (disable) {
                popupDisabled = disable;
            },

            refreshStats: function () {

                //errors count
                elmahr.data.total(
                   _.chain(this.data.allErrors())
                    .reduce(function (acc, e) { return acc + 1; }, 0)
                    .value()
                    );

                for (var k in this.extras) {
                    this.extras[k](this.data.allErrors());
                }
            },

            refreshStatsFor: function (extra) {
                this.extras[extra](this.data.allErrors());
            },

            sortApplications: function (sortedApps) {
                elmahrConnector.sortApplications(sortedApps);
            },

            start: function () {
                ko.applyBindings(this.data);
                this.statusShowUI();
            },

            clearErrors: function (when, predicate, sourceId) {
                var tbra = [],
                    allErrs = this.data.allErrors();

                for (i = 0; i < allErrs.length; i++)
                    if (predicate(allErrs[i].rawTime, when) && allErrs[i].sourceId === sourceId)
                        tbra.push(allErrs[i]);

                for (j = 0; j < tbra.length; j++)
                    this.data.allErrors.remove(tbra[j]);

                this.data.total(elmahr.data.total() - tbra.length >= 0
                                ? elmahr.data.total() - tbra.length
                                : 0);
            },

            advertiseApplication: function (applicationName, infoUrl, sourceId, hasTestExceptionUrl, boxColor, properties, active) {
                var apps = this.data.advertisedApplications(),
                    current = _.chain(apps)
                               .find(function (app) {
                                   return app.sourceId === sourceId;
                               })
                               .value();

                if (!current) {

                    if (!active && !isApplicationBlocked(sourceId))
                        blockedApplications.push(sourceId);
                    else if (active && isApplicationBlocked(sourceId))
                        blockedApplications.splice($.inArray(sourceId, blockedApplications), 1);

                    elmahrConnector.toggleApplicationStatus(sourceId, active);

                    var applicationViewModel = new ApplicationViewModel(
                        applicationName,
                        infoUrl,
                        sourceId,
                        hasTestExceptionUrl,
                        boxColor,
                        this.data.advertisedApplications().length);

                    if (!$.isEmptyObject(this.plugins)) {
                        for (var i in this.plugins) {
                            var plugin = this.plugins[i];
                            plugin(elmahrConnector, applicationViewModel, properties);
                        }
                    }

                    this.data.advertisedApplications.push(applicationViewModel);

                }
            },

            addError: function (envelope, append) {
                var e = new ErrorViewModel(envelope, envelope.SourceId, envelope.InfoUrl);

                var apps = this.data.advertisedApplications();
                for (var a = 0; a < apps.length; a++) {
                    if (apps[a].sourceId === envelope.SourceId) {
                        apps[a].addError(envelope, append);
                    }
                }

                updateLowestIndex(e, append);
                
                if (append)
                    this.data.allErrors.push(e);
                else
                    this.data.allErrors.unshift(e);

                var sourceId = envelope.SourceId,
                    found = _.chain(apps)
                             .find(function (app) {
                                 return app.sourceId === sourceId;
                             })
                             .value();

                if (found) {
                    if (!append) {
                        found.lastError(e);
                        latest(e);
                    }
                    found.errorsCount(found.errorsCount() + 1);

                    if (this.data.unseenErrors !== 'no') {
                        this.data.unseenErrors++;
                        document.title = '(' + this.data.unseenErrors + ') ' + this.config.title;
                    }
                }
            },

            askForMoreErrors: function () {
                elmahr.startTransfer();
                elmahrConnector.askForMoreErrors(lowestIndex);
                return false;
            },

            selectTarget: function (id, sourceId) {
                if (!popupDisabled)
                    elmahrConnector.askFullError(id, sourceId);
            },

            isErrorTypeBlocked: isErrorTypeBlocked,

            isApplicationBlocked: isApplicationBlocked,

            toggleErrorType: function (type, n) {
                if (n === 'pause' && !isErrorTypeBlocked(type))
                    blockedErrorTypes.push(type);
                else if (n === 'play' && isErrorTypeBlocked(type))
                    blockedErrorTypes.splice($.inArray(type, blockedErrorTypes), 1);
            },

            toggleApplication: function (sourceId, status) {
                if (status === pause && !isApplicationBlocked(sourceId)) {
                    blockedApplications.push(sourceId);
                    elmahrConnector.toggleApplicationStatus(sourceId, false);
                } else if (status === play && isApplicationBlocked(sourceId)) {
                    blockedApplications.splice($.inArray(sourceId, blockedApplications), 1);
                    elmahrConnector.toggleApplicationStatus(sourceId, true);
                }
            },

            raiseTestException: function (sourceId) {
                elmahrConnector.raiseTestException(sourceId);
            },

            getErrorsResume: function (sourceId) {
                if (!popupDisabled) {
                    elmahr.onGetErrorsResume(sourceId);
                    elmahrConnector.getErrorsResume(sourceId);
                }
            },

            notifyErrorsResume: function (sourceId, resume) {
                var apps = this.data.advertisedApplications();
                for (var a = 0; a < apps.length; a++) {
                    if (apps[a].sourceId === sourceId) {
                        apps[a].updateErrorsStats(resume);
                    }
                }
            },

            toggleRememberMe: function () {
                var self = this,
                    apps = self.data.advertisedApplications(),
                    ec = elmahrConnector,
                    isBlocked = isApplicationBlocked;
                elmahrConnector.toggleRememberMe().done(function () {
                    for (var a = 0; a < apps.length; a++) {
                        var app = apps[a];
                        ec.toggleApplicationStatus(app.sourceId, !isBlocked(app.sourceId));
                    }
                    ec.sortApplications(self.getSortedApplications());
                });
            },

            toggleLog: function () {
                elmahr.clearLog();
                elmahr.onLog('Subscribing to log...');
                elmahrConnector.toggleLog()
                               .done(function () {
                                   elmahr.onLog('...log ready!');
                               });
            },

            setBrowserVisible: function (visible) {
                this.data.unseenErrors = visible ? 'no' : 0;
                if (visible)
                    document.title = this.config.title;
            },

            askToClearAllErrors: function (sourceId, onFail) {
                askToClear(sourceId, function (app) { app.askToClearAllErrors(onFail); });
            },

            askToClearErrorsBefore: function (sourceId, when, onFail) {
                askToClear(sourceId, function (app) { app.askToClearErrorsBefore(when, onFail); });
            },

            askToClearErrorsAfter: function (sourceId, when, onFail) {
                askToClear(sourceId, function (app) { app.askToClearErrorsAfter(when, onFail); });
            },

            clearAllErrors: function (sourceId, onFail) {
                if (elmahr.config.enableDelete)
                    clearer(sourceId, null, function () { return elmahrConnector.clearAllErrors(sourceId).fail(onFail); }, function (app) { return app.clearAllErrors(); });
                else
                    onFail();
            },

            clearErrorsBefore: function (sourceId, when, onFail) {
                if (elmahr.config.enableDelete)
                    clearer(sourceId, when, function () { return elmahrConnector.clearErrorsBefore(sourceId, when).fail(onFail); }, function (app) { return app.clearErrorsBefore(when); });
                else
                    onFail();
            },

            clearErrorsAfter: function (sourceId, when, onFail) {
                if (elmahr.config.enableDelete)
                    clearer(sourceId, when, function () { return elmahrConnector.clearErrorsAfter(sourceId, when).fail(onFail); }, function (app) { return app.clearErrorsAfter(when); });
                else
                    onFail();
            },

            // init

            init: function () {

                // proxy hub method defs

                $.extend(elmahrConnector, {

                    advertiseApplications: function (apps) {
                        for (var a = 0; a < apps.length; a++) {
                            var app = apps[a];
                            elmahr.advertiseApplication(app.ApplicationName, app.InfoUrl, app.SourceId, app.HasTestExceptionUrl, app.BoxColor, app.Properties, app.Active);
                        }
                    },

                    notifyErrors: function (envelopes, append) {
                        elmahr.statusReceiving();
                        for (var k = 0; k < envelopes.length; k++) {
                            var envelope = envelopes[k];

                            if (!envelope.ApplicationName)
                                continue;   //IE7 bug

                            if (isApplicationBlocked(envelope.SourceId))
                                continue;

                            var type = envelope.Error.ShortType;
                            if (isErrorTypeBlocked(type))
                                continue;

                            elmahr.addError(envelope, append);
                        }

                        if (envelopes) {
                            if (envelopes.length) {
                                elmahr.statusRefreshingStats();
                                elmahr.refreshStats();
                            }
                            else
                                elmahr.hideAskForMore();
                        }

                        if (elmahr.data.selectedApp()) {
                            _.chain(envelopes)
                             .map(function (e) { return e.SourceId; })
                             .uniq()
                             .each(function (id) { elmahr.getErrorsResume(id); });
                        }
                        elmahr.stopTransfer();
                    },

                    notifyFullError: function (error) {
                        if (!error)
                            return;

                        var errorModel = new ErrorViewModel(error);

                        selected(errorModel);
                        elmahr.onSelectError(errorModel);
                    },

                    notifyRememberMe: function (remember) {
                        elmahr.onRememberMe(remember);
                    },

                    notifyErrorsResume: function (sourceId, resume) {
                        elmahr.notifyErrorsResume(sourceId, resume);
                        elmahr.onGetErrorsResumeDone();
                    },

                    log: function (message) {
                        elmahr.onLog(message);
                    }
                });

                // now let's startup things!

                $.connection.hub
                    .starting(function () {
                    })
                    .disconnected(function () {
                        elmahr.showMessage('Hmm, disconnected...', 'Warning');
                    })
                    .reconnected(function () {
                        elmahr.showMessage('Reconnected!');
                    })
                    .error(function (e) {
                        elmahr.showMessage('Oops, an error has occurred... but I will retry!', 'Critical');
                    });

                $.connection(elmahr.config.root + 'commands')
                    .start()
                    .done(function (connection) {
                        elmahr.statusSendingCommand();

                        connection.received(function (answer) {
                            switch (answer.Selector) {
                                case 'startup':
                                    hookPlugins(answer.Answer);

                                    elmahr.statusStartingHubs();
                                    $.connection.hub
                                        .start(function () {
                                            elmahr.statusConnecting();
                                            elmahrConnector.connect().done(function () {
                                                if (elmahr.config.log)
                                                    elmahr.toggleLog();
                                                elmahr.start();
                                                elmahrConnector.retrieveHistoricalErrors().done(function () {
                                                    elmahr.statusConnected();
                                                });
                                            });
                                        });

                                    break;
                            }
                        });

                        connection.send('startup');
                    });

            }

        };

    })();

} ());

